@@ -2,11 +2,22 @@ import logging
 import mmap
 import os
 import re
+import sys
 import shutil
 import threading
-import cPickle as pickle
-import pickle
-from cStringIO import StringIO
+import base64
+
+if sys.version_info[:3] < (3, 0, 0):
+    import cPickle as pickle
+    from cStringIO import StringIO as BytesIO
+    def is_filelike(obj):
+        return isinstance(obj, file)
+else:
+    import pickle
+    from io import BytesIO
+    from io import IOBase
+    def is_filelike(obj):
+        return isinstance(obj, IOBase)
 
 from fcntl import flock
 from fcntl import LOCK_EX
@@ -26,6 +37,7 @@ LOG_STREAM = 2
 re_id = re.compile(r'(?P<protocol>[a-z]+)://(?P<token>.+)')
 logger = logging.getLogger('dobbin.database')
 
+
 class Database(Manager):
     """Object database which stores data in a single file."""
 
@@ -40,8 +52,8 @@ class Database(Manager):
         self._open()
 
         # pickle writer
-        self._buffer = StringIO()
-        self._pickler = pickle.Pickler(self._buffer)
+        self._buffer = BytesIO()
+        self._pickler = pickle.Pickler(self._buffer, pickle.HIGHEST_PROTOCOL)
         self._offsets = {}
 
         super(Database, self).__init__()
@@ -54,6 +66,14 @@ class Database(Manager):
         self._oid = oid
         return oid
 
+    def close(self):
+        self.lock_acquire()
+        try:
+            self._rstream.close()
+            self._rstream = None
+        finally:
+            self.lock_release()
+
     def read(self, jar, timestamp):
         """Read transactions newer than ``timestamp``."""
 
@@ -63,7 +83,9 @@ class Database(Manager):
             try:
                 offset = self._offsets[timestamp]
             except KeyError:
-                raise ValueError("No offset found for timestamp: %s." % timestamp)
+                raise ValueError(
+                    "No offset found for timestamp: %s." % timestamp
+                    )
 
         stream = self._open_mmap(offset)
         if stream is None:
@@ -81,7 +103,8 @@ class Database(Manager):
             token = match.group('token')
 
             if protocol == 'oid':
-                oid, cls = pickle.loads(token.decode('base64'))
+                p = base64.b64decode(token.encode('ascii'))
+                oid, cls = pickle.loads(p)
                 return jar.get(oid, cls)
 
             if protocol == 'file':
@@ -127,21 +150,17 @@ class Database(Manager):
                 if oid is None:
                     oid = self.new_oid(obj)
 
-                return "oid://%s" % pickle.dumps((oid, obj._p_class)).encode('base64').rstrip()
+                p = pickle.dumps((oid, obj._p_class))
+                e = base64.b64encode(p).decode('ascii')
+
+                return "oid://%s" % e
 
             if isinstance(obj, PersistentStream):
                 return "file://%d:%d" % (obj.offset, obj.length)
 
             if isinstance(obj, PersistentFile):
-                # compute file length
-                offset = obj.tell()
-                obj.seek(0, os.SEEK_END)
-                length = obj.tell() - offset
-                obj.seek(offset)
-
                 # write transaction log segment
-                offset = self._write_unbuffered(LOG_STREAM, (obj.name, length))
-                self._write_stream(obj, length)
+                offset, length = self._write_stream(obj)
 
                 # switch identity to transaction stream
                 obj.__dict__.clear()
@@ -150,7 +169,7 @@ class Database(Manager):
 
                 return "file://%d:%d" % (offset, length)
 
-            if isinstance(obj, file):
+            if is_filelike(obj):
                 raise TypeError(
                     "Can't persist files; use the ``PersistentFile`` wrapper.")
 
@@ -173,7 +192,9 @@ class Database(Manager):
                 return
             try:
                 # write transaction record
-                self._write(LOG_RECORD, TransactionRecord(self.tx_timestamp, False))
+                self._write(
+                    LOG_RECORD, TransactionRecord(self.tx_timestamp, False)
+                    )
                 self._flush()
             finally:
                 # update transaction state
@@ -195,16 +216,16 @@ class Database(Manager):
             if self.tx_ref is transaction:
                 return
 
-            self.lock_release()
-
             # open write stream
-            wstream = self._wstream = file(self._path, 'ab+')
+            wstream = self._wstream = open(self._path, 'ab+')
 
             # acquire commit lock
             fd = wstream.fileno()
             self._commitlock_acquire = lambda: flock(fd, LOCK_EX | LOCK_NB)
             self._commitlock_release = lambda: flock(fd, LOCK_UN)
 
+            self.lock_release()
+
             try:
                 self._commitlock_acquire()
             except IOError:
@@ -242,7 +263,9 @@ class Database(Manager):
                 return
             try:
                 # write transaction record
-                self._write(LOG_RECORD, TransactionRecord(self.tx_timestamp, True))
+                self._write(
+                    LOG_RECORD, TransactionRecord(self.tx_timestamp, True)
+                    )
                 self._flush()
             finally:
                 # update transaction state
@@ -278,7 +301,7 @@ class Database(Manager):
 
     def _open(self):
         if os.path.exists(self._path):
-            f = self._rstream = file(self._path, 'rb+')
+            f = self._rstream = open(self._path, 'rb+')
             return f
 
     def _open_mmap(self, offset=0):
@@ -293,44 +316,50 @@ class Database(Manager):
         return _map
 
     def _opener(self):
-        return file(self._path, 'rb')
+        return open(self._path, 'rb')
 
     def _write(self, segment_type, data):
         try:
             self._pickler.dump((segment_type, data))
-        except Exception, e:
+        except Exception as e:
             logger.critical("Could not pickle data: %s (type %d).\n%s" % (
                 repr(data), segment_type, str(e)))
             self._flush()
             raise
 
-    def _write_unbuffered(self, segment_type, data):
-        offset = self._buffer.tell()
-        self._write(segment_type, data)
-        return self._flush(offset)
-
-    def _write_stream(self, stream, length):
+    def _write_stream(self, stream):
+        pos = stream.tell()
+        stream.seek(0, os.SEEK_END)
+        length = stream.tell() - pos
+        stream.seek(pos)
+        log = pickle.dumps((LOG_STREAM, (stream.name, length)))
+        self._wstream.write(log)
+        offset = self._wstream.tell()
         shutil.copyfileobj(stream, self._wstream, length)
+        stream.close()
+        return offset, length
+
 
 class TransactionRecord(object):
     def __init__(self, timestamp, status):
         self.timestamp = timestamp
         self.status = status
 
-class PersistentStream(threading.local):
+
+class PersistentStream(object):
     """Binary stream persisted in the transaction log.
 
     Features a file-like API as well as iteration (independent from
     each other; iteration will always acquire its own file handle).
     """
 
-    stream = None
     chunk_size = 32768
 
     def __init__(self, opener, offset, length):
-        self._opener = opener
         self.offset = offset
         self.length = length
+        self._opener = opener
+        self._thread = threading.local()
 
     def __deepcopy__(self, memo):
         return self
@@ -353,14 +382,26 @@ class PersistentStream(threading.local):
 
         while remaining > 0:
             count = min(chunk_size, remaining)
-            bytes = read(remaining, f)
+            bytes = read(count, f)
             remaining -= len(bytes)
             yield bytes
 
+        f.close()
+
+    def _get_thread_local_stream(self):
+        try:
+            return self._thread.stream
+        except AttributeError:
+            return
+
+    def _set_thread_local_stream(self, stream):
+        self._thread.stream = stream
+
+    stream = property(_get_thread_local_stream, _set_thread_local_stream)
+
     @property
     def closed(self):
-        stream = self.stream
-        if stream is None:
+        if self.stream is None:
             return True
         return self.stream.closed
 
